/*
   CIPEX - encrypted and compressed IP over UDP tunneling

   mlzocipe.c - the compression/expansion part of the CIPEX device

   Copyright 2003 J.J.G. Steegers <hsx@dds.nl>
    
   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version
   2 of the License, or (at your option) any later version.
*/
#include <net/ip.h>
#include <net/tcp.h>
#include <linux/if_arp.h>
#include <linux/string.h>
#include "cipe.h"
#ifdef MLZO_CIPE
# include "mlzoprot.h"

static char *secs2dhms(unsigned long lseconds) {
    int days, hours, minutes, seconds;
    unsigned long int   rest;
    static char outstring[32];
    days=(int)(lseconds / 86400);
    rest=(lseconds % 86400);
    hours=(int)(rest / 3600);
    rest=(rest % 3600);
    minutes=(int)(rest / 60);
    rest=(rest % 60);
    seconds=(int)rest;
    sprintf(outstring,"%4d d, %2d h, %2d m, %2d s",
	    days,hours,minutes,seconds);
    return (char *) outstring;
}

void cipex_clearstats(struct cipe *c) {
    c->rx_uncomp_pkts=0L;
    c->rx_uncomp_bytes=0L;
    c->rx_comp_pkts=0L;
    c->rx_comp_rbytes=0L;
    c->rx_comp_cbytes=0L;
    c->tx_uncomp_pkts=0L;
    c->tx_uncomp_bytes=0L;
    c->tx_comp_pkts=0L;
    c->tx_comp_rbytes=0L;
    c->tx_comp_cbytes=0L;
}

void  cipex_stats(struct cipe *c){
    unsigned long compression_ratio=0L;
    unsigned long rx_tot_raw, tx_tot_raw; 
    unsigned long rx_tot_tnl, tx_tot_tnl; 
    printk(KERN_INFO "%s: CipeX vers. %s statistics:\n",  
        c->dev->name, VERSION	);
    printk(KERN_INFO "%s: RX uncompr. packets : %12lu \n",
        c->dev->name, c->rx_uncomp_pkts);
    printk(KERN_INFO "%s: RX uncompr. bytes   : %12lu \n",
        c->dev->name, c->rx_uncomp_bytes);
    printk(KERN_INFO "%s: RX compr. packets   : %12lu \n",
        c->dev->name, c->rx_comp_pkts);
    printk(KERN_INFO "%s: RX compr. raw bytes : %12lu \n",
        c->dev->name, c->rx_comp_rbytes);
    printk(KERN_INFO "%s: RX compr. rx bytes  : %12lu \n",
        c->dev->name, c->rx_comp_cbytes);
    if (c->rx_comp_cbytes > 0L) { compression_ratio =
	    (100L*c->rx_comp_rbytes)/c->rx_comp_cbytes;
    } else compression_ratio = 100L;
    printk(KERN_INFO "%s: RX compression ratio: %9lu.%02lu \n",
        c->dev->name,compression_ratio/100L,compression_ratio%100L);
    rx_tot_tnl = c->rx_comp_cbytes + c->rx_uncomp_bytes ;
    rx_tot_raw = c->rx_comp_rbytes + c->rx_uncomp_bytes ;
    printk(KERN_INFO "%s: RX total rx bytes   : %12lu \n",
	    c->dev->name, rx_tot_tnl );
    printk(KERN_INFO "%s: RX total raw bytes  : %12lu \n",
            c->dev->name, rx_tot_raw );
    if (rx_tot_tnl  > 0L) 
        compression_ratio = (100L * rx_tot_raw) / rx_tot_tnl;
    else
	compression_ratio = 100L;
    printk(KERN_INFO "%s: RX compression total: %9lu.%02lu \n",
        c->dev->name, compression_ratio/100L, compression_ratio%100L);

    printk(KERN_INFO "%s: TX uncompr. packets : %12lu \n",
    c->dev->name, c->tx_uncomp_pkts);
    printk(KERN_INFO "%s: TX uncompr. bytes   : %12lu \n",
    c->dev->name, c->tx_uncomp_bytes);
    printk(KERN_INFO "%s: TX compr. packets   : %12lu \n",
	c->dev->name, c->tx_comp_pkts);
    printk(KERN_INFO "%s: TX compr. raw bytes : %12lu \n",
	c->dev->name, c->tx_comp_rbytes);
    printk(KERN_INFO "%s: TX compr. tx bytes  : %12lu \n",
	c->dev->name, c->tx_comp_cbytes);
    if (c->tx_comp_cbytes > 0L) { compression_ratio = 
	(100L * c->tx_comp_rbytes) / c->tx_comp_cbytes;
    } else compression_ratio = 100L;
    printk(KERN_INFO "%s: TX compression ratio: %9lu.%02lu \n",
        c->dev->name, compression_ratio/100L, compression_ratio%100L);
    tx_tot_tnl = c->tx_comp_cbytes  + c->tx_uncomp_bytes ;
    tx_tot_raw = c->tx_comp_rbytes  + c->tx_uncomp_bytes ;
    printk(KERN_INFO "%s: TX total rx bytes   : %12lu \n",
        c->dev->name, tx_tot_tnl );
    printk(KERN_INFO "%s: TX total raw bytes  : %12lu \n",
        c->dev->name, tx_tot_raw );
    if (tx_tot_tnl > 0L) 
	compression_ratio = (100L * tx_tot_raw) / tx_tot_tnl;
    else
	compression_ratio = 100L;
    printk(KERN_INFO "%s: TX compression total: %9lu.%02lu \n",
        c->dev->name, compression_ratio/100L, compression_ratio%100L);
    if ( c-> nocomp ) {
        printk(KERN_INFO "%s: TX compr. disabled!  \n", c->dev->name );
    } else {
        printk(KERN_INFO "%s: TX min. packet size : %12d \n",
	    c->dev->name, c->minsizecomp);
    }
    printk(KERN_INFO "%s:%010lu=%24s\n", c->dev->name,
	(CURRENT_TIME - c->timestamp_load),
         secs2dhms(CURRENT_TIME - c->timestamp_load) );
}

int cipex_compress(struct cipe *c, unsigned char *buffer, int ilen  ) {
    unsigned char* lzo_wrk;
    unsigned char* lzo_buf;
    int olen; 

    dprintk(DEB_MLZO, (KERN_DEBUG 
	"[%s]cipe_compress: ilen=%d ms=%d comp=%c.\n",
	c->dev->name, ilen, c->minsizecomp,
	c->nocomp ? 'd' : 'e' ));

    if ((ilen < c->minsizecomp)|| (c->nocomp) ){
	dprintk(DEB_MLZO, (KERN_DEBUG 
		"[%s]cipe_compress: no compr. ilen=%d (< ms=%d)\n",
		c->dev->name, ilen, c->minsizecomp	));
	return 0;
    }
    lzo_wrk = c->lzo_wrk; 
    lzo_buf = c->lzo_buf; 

    if (mlzo_compress(buffer,lzo_buf,lzo_wrk,ilen,&olen)){
	dprintk(DEB_MLZO,(KERN_DEBUG 
		"[%s]cipe_compress: ilen=%d -> olen=%d\n",
		c->dev->name, ilen, olen));
	if ( ilen > olen ){ /* is compressed */
#	ifdef DEBUG
	    if (cipe_debug&DEB_MLZO){
		LOCK_PRINTK;
		printk(KERN_DEBUG 
		    "[%s]cipe_compress: olen=%d hexdump of buf:\n",
		    c->dev->name, olen);
		cipe_hexdump(lzo_buf,olen);
		UNLOCK_PRINTK;
	    }
#	endif
	    memcpy(buffer,lzo_buf,olen); 
	    return(olen); 
	}
    }
    return(0); 
} 

int cipex_expand(struct cipe *c, struct sk_buff **skbufp, int *len) {
    unsigned char* lzo_buf;
    int olen ,ilen;
    unsigned char *buffer = (*skbufp)->data;

#ifdef DEBUG
    if (cipe_debug&DEB_MLZO){
        LOCK_PRINTK;
        printk(KERN_DEBUG 
		"[%s]cipe_expand: len=%d - hexdump of buffer at %p\n", 
			c->dev->name, *len, buffer);
			cipe_hexdump(buffer,*len, buffer);
        UNLOCK_PRINTK;
    }
#endif

    ilen = (*len) ; 
    lzo_buf = c->lzo_buf;

    if (ilen > 0) {
	if (mlzo_expand(buffer,lzo_buf,ilen,&olen)){
#	ifdef DEBUG
	    if (cipe_debug&DEB_MLZO){
		LOCK_PRINTK;
		printk(KERN_DEBUG 
		    "[%s]cipe_expand: %d -> %d - hexdump of lzobuf:\n", 
		    c->dev->name, ilen, olen);
		cipe_hexdump(lzo_buf,olen);
		UNLOCK_PRINTK;
	    }
#	endif
	    dprintk(DEB_MLZO, (KERN_DEBUG 
    		"[%s]cipe_expand: OLD SKB at %p: len=%d hr= %d, tr=%d\n", 
        	c->dev->name,buffer,(*skbufp)->len,
		skb_headroom(*skbufp), skb_tailroom(*skbufp)));

            (*skbufp)->len = ilen;
	    if (olen < ilen){ /* expand program error */
		    printk(KERN_ERR 
    		    "[%s]cipe_expand: length problem: olen=%d < len=%d\n",
		    	(*skbufp)->dev->name, olen, ilen);
		     return (0);
	    }

	    {
		struct sk_buff *x ;
	    	x = skb_copy_expand(*skbufp,8,(olen-ilen),GFP_ATOMIC); 
                if (!x) {
		    printk(KERN_ERR 
    		    "[%s]cipe_expand: Out of memory, expand failed\n",
		    	(*skbufp)->dev->name);
		    return (0);
	        }
    		if ( (*skbufp)->sk )
		    skb_set_owner_w(x, (*skbufp)->sk);

    		dev_kfree_skb(*skbufp); 
		*skbufp = x; 
	    }
	    
	    (*skbufp)->len -= ilen;
    	    (*skbufp)->tail -= ilen;
	    skb_put(*skbufp, olen );
    	    buffer = (*skbufp)->data;
	    memcpy(buffer, lzo_buf, olen); 
    	    *len = olen;

	    dprintk(DEB_MLZO, (KERN_DEBUG 
    		"[%s]cipe_expand: NEW SKB at %p: len=%d hr= %d, tr=%d\n", 
                c->dev->name, buffer,(*skbufp)->len,
		skb_headroom(*skbufp), skb_tailroom(*skbufp)));

	    return(1);
	} 
    } 
    return(0);
} 
#endif /* MLZO_CIPE */
